@@ -4,7 +4,7 @@
 """
 ===============================================================================
 
-    Copyright (C) 2009-2018 Rudolf Cardinal (rudolf@pobox.com).
+    Original code copyright (C) 2009-2018 Rudolf Cardinal (rudolf@pobox.com).
 
     This file is part of cardinal_pythonlib.
 
@@ -21,6 +21,9 @@
     limitations under the License.
 
 ===============================================================================
+
+**Constants for use with HTTP.**
+
 """
 
 
@@ -29,6 +32,8 @@ class MimeType(object):
     Some MIME type constants.
     See also the Python standard library 'mimetypes'; e.g.
 
+    .. code-block:: python
+
         import mimetypes
         mimetypes.types_map['.pdf']  # 'application/pdf'
     """
