@@ -4,7 +4,7 @@
 """
 ===============================================================================
 
-    Copyright (C) 2009-2018 Rudolf Cardinal (rudolf@pobox.com).
+    Original code copyright (C) 2009-2018 Rudolf Cardinal (rudolf@pobox.com).
 
     This file is part of cardinal_pythonlib.
 
@@ -21,7 +21,29 @@
     limitations under the License.
 
 ===============================================================================
-"""
+
+**Add "INSERT ON DUPLICATE KEY UPDATE" functionality to SQLAlchemy for MySQL.**
+
+- https://www.reddit.com/r/Python/comments/p5grh/sqlalchemy_whats_the_idiomatic_way_of_writing/
+- https://github.com/bedwards/sqlalchemy_mysql_ext/blob/master/duplicate.py
+  ... modified
+- http://docs.sqlalchemy.org/en/rel_1_0/core/compiler.html
+- http://stackoverflow.com/questions/6611563/sqlalchemy-on-duplicate-key-update
+- http://dev.mysql.com/doc/refman/5.7/en/insert-on-duplicate.html
+
+Once implemented, you can do
+
+.. code-block:: python
+
+    q = sqla_table.insert_on_duplicate().values(destvalues)
+    session.execute(q)
+
+**Note: superseded by SQLAlchemy v1.2:**
+
+- http://docs.sqlalchemy.org/en/latest/changelog/migration_12.html
+- http://docs.sqlalchemy.org/en/latest/dialects/mysql.html#mysql-insert-on-duplicate-key-update
+
+"""  # noqa
 
 import logging
 import re
@@ -37,29 +59,12 @@ log = logging.getLogger(__name__)
 log.addHandler(logging.NullHandler())
 
 
-# =============================================================================
-# INSERT ... ON DUPLICATE KEY UPDATE support, for MySQL
-# =============================================================================
-# https://www.reddit.com/r/Python/comments/p5grh/sqlalchemy_whats_the_idiomatic_way_of_writing/  # noqa
-# https://github.com/bedwards/sqlalchemy_mysql_ext/blob/master/duplicate.py
-# ... modified
-# http://docs.sqlalchemy.org/en/rel_1_0/core/compiler.html
-# http://stackoverflow.com/questions/6611563/sqlalchemy-on-duplicate-key-update
-# http://dev.mysql.com/doc/refman/5.7/en/insert-on-duplicate.html
-#
-# Once implemented, you can do
-#       q = sqla_table.insert_on_duplicate().values(destvalues)
-#       session.execute(q)
-
-# =============================================================================
-# NOTE: SQLALCHEMY SUPPORTS THIS NATIVELY AS OF V1.2:
-# =============================================================================
-# http://docs.sqlalchemy.org/en/latest/changelog/migration_12.html
-# http://docs.sqlalchemy.org/en/latest/dialects/mysql.html#mysql-insert-on-duplicate-key-update  # noqa
-
-
 # noinspection PyAbstractClass
 class InsertOnDuplicate(Insert):
+    """
+    Class that derives from :class:`Insert`, so we can hook in to operations
+    involving it.
+    """
     pass
 
 
@@ -67,11 +72,30 @@ def insert_on_duplicate(tablename: str,
                         values: Any = None,
                         inline: bool = False,
                         **kwargs):
+    """
+    Command to produce an :class:`InsertOnDuplicate` object.
+
+    Args:
+        tablename: name of the table
+        values: values to ``INSERT``
+        inline: as per 
+            http://docs.sqlalchemy.org/en/latest/core/dml.html#sqlalchemy.sql.expression.insert
+        kwargs: additional parameters
+
+    Returns:
+        an :class:`InsertOnDuplicate` object
+
+    """  # noqa
     return InsertOnDuplicate(tablename, values, inline=inline, **kwargs)
 
 
 # noinspection PyPep8Naming
 def monkeypatch_TableClause() -> None:
+    """
+    Modifies :class:`sqlalchemy.sql.expression.TableClause` to insert
+    a ``insert_on_duplicate`` member that is our :func:`insert_on_duplicate`
+    function as above.
+    """
     log.debug("Adding 'INSERT ON DUPLICATE KEY UPDATE' support for MySQL "
               "to SQLAlchemy")
     TableClause.insert_on_duplicate = insert_on_duplicate
@@ -79,6 +103,9 @@ def monkeypatch_TableClause() -> None:
 
 # noinspection PyPep8Naming
 def unmonkeypatch_TableClause() -> None:
+    """
+    Reverses the action of :func:`monkeypatch_TableClause`.
+    """
     del TableClause.insert_on_duplicate
 
 
@@ -98,16 +125,22 @@ def compile_insert_on_duplicate_key_update(insert: Insert,
                                            compiler: SQLCompiler,
                                            **kw) -> str:
     """
-    We can't get the fieldnames directly from 'insert' or 'compiler'.
-    We could rewrite the innards of the visit_insert statement, like
-        https://github.com/bedwards/sqlalchemy_mysql_ext/blob/master/duplicate.py  # noqa
-    ... but, like that, it will get outdated.
-    We could use a hack-in-by-hand method, like
-        http://stackoverflow.com/questions/6611563/sqlalchemy-on-duplicate-key-update
-    ... but a little automation would be nice.
-    So, regex to the rescue.
-    NOTE THAT COLUMNS ARE ALREADY QUOTED by this stage; no need to repeat.
-    """
+    Hooks into the use of the :class:`InsertOnDuplicate` class
+    for the MySQL dialect. Compiles the relevant SQL for an ``INSERT...
+    ON DUPLICATE KEY UPDATE`` statement.
+
+    Notes:
+ 
+    - We can't get the fieldnames directly from ``insert`` or ``compiler``.
+    - We could rewrite the innards of the visit_insert statement
+      (https://github.com/bedwards/sqlalchemy_mysql_ext/blob/master/duplicate.py)... 
+      but, like that, it will get outdated.
+    - We could use a hack-in-by-hand method
+      (http://stackoverflow.com/questions/6611563/sqlalchemy-on-duplicate-key-update)
+      ... but a little automation would be nice.
+    - So, regex to the rescue.
+    - NOTE THAT COLUMNS ARE ALREADY QUOTED by this stage; no need to repeat.
+    """  # noqa
     # log.critical(compiler.__dict__)
     # log.critical(compiler.dialect.__dict__)
     # log.critical(insert.__dict__)
